/*
 * auto_explain.c
 */
#include "postgres.h"

#include "commands/explain.h"
#include "executor/instrument.h"
#include "miscadmin.h"
#include "utils/guc_tables.h"

PG_MODULE_MAGIC;

#define GUCNAME(name)		("explain." name)

static int		explain_log_min_duration = -1;	/* msec or -1 */
static bool		explain_log_analyze = false;
static bool		explain_log_verbose = false;

static bool		toplevel = true;
static ExecutorRun_hook_type	prev_ExecutorRun = NULL;

void	_PG_init(void);
void	_PG_fini(void);

static TupleTableSlot *explain_ExecutorRun(QueryDesc *queryDesc,
										   ScanDirection direction,
										   long count);
static bool assign_log_min_duration(int newval, bool doit, GucSource source);
static bool assign_log_analyze(bool newval, bool doit, GucSource source);
static bool assign_log_verbose(bool newval, bool doit, GucSource source);

static struct config_int def_log_min_duration =
{
	{
		GUCNAME("log_min_duration"),
		PGC_USERSET,
		STATS_MONITORING,
		"Sets the minimum execution time above which plans will be logged.",
		"Zero prints all plans. -1 turns this feature off.",
		GUC_UNIT_MS
	},
	&explain_log_min_duration,
	-1, -1, INT_MAX / 1000, assign_log_min_duration, NULL
};

static struct config_bool def_log_analyze =
{
	{
		GUCNAME("log_analyze"),
		PGC_USERSET,
		STATS_MONITORING,
		"Use EXPLAIN ANALYZE for plan logging."
	},
	&explain_log_analyze,
	false, assign_log_analyze, NULL
};

static struct config_bool def_log_verbose =
{
	{
		GUCNAME("log_verbose"),
		PGC_USERSET,
		STATS_MONITORING,
		"Use EXPLAIN VERBOSE for plan logging."
	},
	&explain_log_verbose,
	false, assign_log_verbose, NULL
};

void
_PG_init(void)
{
	DefineCustomVariable(PGC_INT, &def_log_min_duration);
	DefineCustomVariable(PGC_BOOL, &def_log_analyze);
	DefineCustomVariable(PGC_BOOL, &def_log_verbose);

	/* install ExecutorRun_hook */
	prev_ExecutorRun = ExecutorRun_hook;
	ExecutorRun_hook = explain_ExecutorRun;
}

void
_PG_fini(void)
{
	/* uninstall ExecutorRun_hook */
	ExecutorRun_hook = prev_ExecutorRun;
}

TupleTableSlot *
explain_ExecutorRun(QueryDesc *queryDesc, ScanDirection direction, long count)
{
	TupleTableSlot *result;
	
	if (toplevel && explain_log_min_duration >= 0)
	{
		instr_time		starttime;
		instr_time		duration;
		double			msec;

		/* Disable our hooks temporarily during the top-level query. */
		toplevel = false;
		PG_TRY();
		{
			INSTR_TIME_SET_CURRENT(starttime);

			if (prev_ExecutorRun)
				result = prev_ExecutorRun(queryDesc, direction, count);
			else
				result = standard_ExecutorRun(queryDesc, direction, count);

			INSTR_TIME_SET_CURRENT(duration);
			INSTR_TIME_SUBTRACT(duration, starttime);
			msec = INSTR_TIME_GET_MILLISEC(duration);

			/* Log plan if duration is exceeded. */
			if (msec > explain_log_min_duration)
			{
				StringInfoData	buf;

				initStringInfo(&buf);
				ExplainOneResult(&buf, queryDesc,
					queryDesc->doInstrument && explain_log_analyze,
					explain_log_verbose);

				/* Remove last line break */
				if (buf.len > 0 && buf.data[buf.len - 1] == '\n')
				{
					buf.data[buf.len - 1] = '\0';
					buf.len--;
				}
				ereport(LOG,
						(errmsg("duration: %.3f ms  plan:\n%s",
								msec, buf.data)));

				pfree(buf.data);
			}

			toplevel = true;
		}
		PG_CATCH();
		{
			toplevel = true;
			PG_RE_THROW();
		}
		PG_END_TRY();
	}
	else
	{
		/* ignore recursive executions, that are typically function calls */
		if (prev_ExecutorRun)
			result = prev_ExecutorRun(queryDesc, direction, count);
		else
			result = standard_ExecutorRun(queryDesc, direction, count);
	}

	return result;
}

/* Emulate PGC_SUSET for custom variables. */
static bool
suset_assign(GucSource source, const char *name)
{
	if (source >= PGC_S_CLIENT && !superuser())
	{
		ereport(GUC_complaint_elevel(source),
				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
				 errmsg("permission denied to set parameter \"%s\"", name)));
		return false;
	}

	return true;
}

static bool
assign_log_min_duration(int newval, bool doit, GucSource source)
{
	if (!suset_assign(source, GUCNAME("log_min_duration")))
		return false;
	if (doit)
		force_instrument = (newval >= 0);
	return true;
}

static bool
assign_log_analyze(bool newval, bool doit, GucSource source)
{
	return suset_assign(source, GUCNAME("log_analyze"));
}

static bool
assign_log_verbose(bool newval, bool doit, GucSource source)
{
	return suset_assign(source, GUCNAME("log_verbose"));
}
